#include "color.frag.dxil.h"
#include "linepoint.vert.dxil.h"
#include "texture_advanced.frag.dxil.h"
#include "texture_rgb.frag.dxil.h"
#include "texture_rgba.frag.dxil.h"
#include "tri_color.vert.dxil.h"
#include "tri_texture.vert.dxil.h"
